#pragma OPENCL EXTENSION cl_khr_fp64: enable
#define BLOCK_SIZE 128
#define WARP_SIZE 32

__kernel void cfft4_initU_kernel(int nr_threads, double ur0, __global double *ucompr, __global double *ucompi)
{
  const unsigned long thread_id = get_global_id(0) + 2;
  const double pi = 3.141592653589793;

  int j;

  double t, ti;

  if (get_local_id(0) == 0) {
    ucompr[0] = ur0;
    ucompi[0] = 0.0;
  }

  if (get_local_id(0) < 2) {
    ucompi[get_local_id(0) + 1] = 0.0;
    ucompr[get_local_id(0) + 1] = 0.0;
  }
  ucompr[1] = 1.0;

  barrier(CLK_GLOBAL_MEM_FENCE);

  for( j = thread_id; j < (nr_threads + 2); j+= (get_num_groups(0) * get_local_size(0)))
  {
    int my_j = ceil(log2((double)((j<<1)+j+1))/2-1);

    int my_ln = 1<<(my_j<<1);
    int my_ku = (2<<((((my_j-1)+1))<<1)-1)/3 + 1;

    int my_i = j-my_ku;

    t  = pi/(my_ln<<1);
    ti = my_i * t;

    ucompr[j] = cos(ti);
    ucompi[j] = sin(ti);

  }
}

__kernel void cfft4_opencl_r1_kernel(int is, unsigned long n,
                                     __global double *xrcomp, __global double *xicomp,
                                     __global double *wrcomp, __global double *wicomp)
{
  const int gidx = get_global_id(0);

  int eightidx = gidx & 7;
  int eightnr = get_local_id(0) >> 3;

  int j, i;

  int n_zwi = n>>2;

  int ln = n_zwi;

  __local int v1[8];      // 1, 1, 1, 1, 1, 1,-1, 1
  __local int v2[8];      // 1, 1,-1,-1, 1, 1, 1,-1

  if (get_local_id(0) < 8) {
    v1[get_local_id(0)] = 1;
    v2[get_local_id(0)] = 1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) < 2 ) {
    v2[2 + (get_local_id(0) & 1)] = -1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) == 0) {
    v1[6] = -1 * is;
    v1[7] *= is;

    v2[6] *= is;
    v2[7] = -1 * is;
  }

  // this sync is crucial with WARP_SIZE > 32
  //barrier(CLK_LOCAL_MEM_FENCE);

  __local double c[BLOCK_SIZE];

  __local double d[BLOCK_SIZE];

  __local double w_zwi[BLOCK_SIZE];


  int x0 = eightidx&1;                  // 0 1 0 1 0 1 0 1
  int x1 = 1-x0;                        // 1 0 1 0 1 0 1 0
  int x3 = eightidx>>2;                 // 0 0 0 0 1 1 1 1
  int x2 = (eightidx>>1)&x3;            // 0 0 0 0 0 0 1 1

  int f1 = x0+(x3<<1)+x2*(-1+(x1<<1));  // 0 1 0 1 2 3 3 2
  int f2 = f1 + 4;                      // 4 5 4 5 6 7 7 6

  int w_1 = 1 - ((eightidx >> 2) << 1);
  int baseval = eightnr<<3;

  for( j = gidx>>3; j <=ln - 1; j+=(get_num_groups(0) * get_local_size(0))>>3 )
  {
    int j4  = 4*j;

    if (eightidx < 4)
    {
      i = (eightnr<<2)+eightidx;
      c[2 * i]     = xrcomp[j+(eightidx*n_zwi)];
      c[2 * i + 1] = xicomp[j+(eightidx*n_zwi)];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    double a8 = c[f1+baseval];
    double b8 = c[f2+baseval];

    d[get_local_id(0)] = v1[eightidx]*a8 + v2[eightidx]*b8;

    barrier(CLK_LOCAL_MEM_FENCE);

    w_zwi[get_local_id(0)] = d[(eightidx&3)+baseval] + w_1*d[(eightidx&3)+4+baseval];

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4) {
     i = (eightnr<<2)+eightidx;
     wrcomp[j4+eightidx] = w_zwi[2 * i];
     wicomp[j4+eightidx] = w_zwi[2 * i + 1];
    }
  }
}

__kernel void cfft4_opencl_r2_kernel(int is, int ln, int n,
                            __global double *urcomp, __global double *uicomp,
                            __global double *xrcomp, __global double *xicomp,
                            __global double *wrcomp, __global double *wicomp)
{

  const unsigned int gidx = get_global_id(0);

  int eightidx = gidx & 7;

  int eightnr = (get_local_id(0) >> 3);

  int kk = get_group_id(1);

  int jj, i;
  int j4,j5;

  double tmp1;
  double tmp2;
  double tmp3;

  __local int j[4];
  __local int k[4];

  if (get_local_id(0) < 4) {
    j[get_local_id(0)] = kk + ((get_local_id(0))*4);
  }

  if (get_local_id(0) < 4) {
    k[get_local_id(0)] = kk + (get_local_id(0) * (n/4));
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  __local double u[8];

  if (get_local_id(0) == 0)
  {
    u[0] = 1.0;
    u[1] = 0.0;

    double urtmp = urcomp[kk+2];
    double uitmp = uicomp[kk+2];

    u[2] = urtmp;
    u[3] = is*uitmp;

    u[4] = u[2]*u[2] - u[3]*u[3];
    u[5] = 2.0 * u[2]*u[3];
    u[6] = u[2]*u[4] - u[3]*u[5];
    u[7] = u[2]*u[5] + u[3]*u[4];
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  __local int v1[8]; // = {1, 1, 1, 1, 1, 1,-1, 1};
  __local int v2[8]; // = {1, 1,-1,-1, 1, 1, 1,-1};

  if (get_local_id(0) < 8) {
    v1[get_local_id(0)] = 1;
    v2[get_local_id(0)] = 1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) < 2 ) {
    v2[2 + (get_local_id(0) & 1)] = -1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) == 0) {
    v1[6] = -1 * is;
    v1[7] *= is;

    v2[6] *= is;
    v2[7] = -1 * is;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  int v3 = 1 - ((eightidx >> 2) << 1);

  int myafak = eightidx>>((eightidx>>2)+1);

  __local double c[BLOCK_SIZE];
  __local double d[BLOCK_SIZE];

  __local double w_zwi[BLOCK_SIZE];
  __local double x_zwi[BLOCK_SIZE];

  int x0 = eightidx&1;           // 0 1 0 1 0 1 0 1
  int x1 = 1-x0;                 // 1 0 1 0 1 0 1 0
  int x3 = eightidx>>2;          // 0 0 0 0 1 1 1 1
  int x2 = (eightidx>>1)&x3;     // 0 0 0 0 0 0 1 1

  int f1 = x0+(x3<<1)+x2*(-1+(x1<<1));  // 0 1 0 1 2 3 3 2
  int f2 = f1 + 4;                      // 4 5 4 5 6 7 7 6

  for( jj = gidx/8; jj <= ln - 1; jj+=get_num_groups(0)*get_local_size(0)/8 )
  {

    j4  = 4*jj;

    if (eightidx < 4)
    {
      i = eightnr*4+eightidx;
      w_zwi[2 * i]     = wrcomp[j4+k[eightidx]];
      w_zwi[2 * i + 1] = wicomp[j4+k[eightidx]];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    int wi_0 = ((eightidx >> 1) << 1);
    int wi_1 = wi_0 + 1;
    int alt_fak = (-1 + ((eightidx & 1)<<1)) * myafak; //afak2[eightidx];

    int www = ((eightidx / 2) * 2) + (1 - (eightidx % 2));

    c[get_local_id(0)] = u[wi_0]*w_zwi[eightidx+(eightnr<<3)] + alt_fak*u[wi_1]*w_zwi[www+(eightnr<<3)];

    barrier(CLK_LOCAL_MEM_FENCE);

    d[get_local_id(0)] = v1[eightidx]*c[f1+(eightnr<<3)] + v2[eightidx]*c[f2+(eightnr<<3)];

    barrier(CLK_LOCAL_MEM_FENCE);

    j5  = 16*jj;

    x_zwi[get_local_id(0)] = d[(eightidx%4)+(eightnr<<3)] + v3*d[eightidx%4+4+(eightnr<<3)];

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4)
    {
      i = eightnr*4+eightidx;
      xrcomp[j5+j[eightidx]] = x_zwi[2 * i];
      xicomp[j5+j[eightidx]] = x_zwi[2 * i + 1];
    }
  }
}

__kernel void cfft4_opencl_r3_kernel(int is, int n, int ln, int ku, int l_in,
                                     __global double *urcomp, __global double *uicomp,
                                     __global double *xrcomp, __global double *xicomp,
                                     __global double *wrcomp, __global double *wicomp)
{

  const unsigned int gidx = get_local_size(0)*get_group_id(0)+get_local_id(0);

  int eightidx = gidx & 7;

  int eightnr = (get_local_id(0) >> 3); 
  int i; 
  int l1 = l_in;

  __local int v1[8]; // = {1, 1, 1, 1, 1, 1,-1, 1};
  __local int v2[8]; // = {1, 1,-1,-1, 1, 1, 1,-1};

  if (get_local_id(0) < 8) {
    v1[get_local_id(0)] = 1;
    v2[get_local_id(0)] = 1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) < 2 ) {
    v2[2 + (get_local_id(0) & 1)] = -1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) == 0) {
    v1[6] = -1 * is;
    v1[7] *= is;

    v2[6] *= is;
    v2[7] = -1 * is;
  }
  barrier(CLK_LOCAL_MEM_FENCE);

  int x0 = eightidx&1;           // 0 1 0 1 0 1 0 1
  int x1 = 1-x0;                 // 1 0 1 0 1 0 1 0
  int x3 = eightidx>>2;          // 0 0 0 0 1 1 1 1
  int x2 = (eightidx>>1)&x3;     // 0 0 0 0 0 0 1 1

  int f1 = x0+(x3<<1)+x2*(-1+(x1<<1));  // 0 1 0 1 2 3 3 2
  int f2 = f1 + 4;                      // 4 5 4 5 6 7 7 6    

  int jj;
  for (jj=get_group_id(1); jj<ln; jj+=get_num_groups(1))
  {
    int kk;

    int j_tmp = (jj*l1*4) + l1*(eightidx/2);
    int k_zwi = (jj*l1) + (eightidx * (n/4)); 

    __local double x_zwi[BLOCK_SIZE];

    __local double c[BLOCK_SIZE];
    __local double d[BLOCK_SIZE];
    __local double u[BLOCK_SIZE];

    __local double w_zwi[BLOCK_SIZE];

    for( kk = gidx/8; kk <= l1 - 1; kk+=get_num_groups(0)*get_local_size(0)/8 )
    {
      if (eightidx == 0)
      {
        u[get_local_id(0)  ] = 1.0;
        u[get_local_id(0)+1] = 0.0;
        u[get_local_id(0)+2] = urcomp[kk+ku-1];
        u[get_local_id(0)+3] = uicomp[kk+ku-1];
        u[get_local_id(0)+4] = u[get_local_id(0)+2] * u[get_local_id(0)+2]-u[get_local_id(0)+3] * u[get_local_id(0)+3];
        u[get_local_id(0)+5] = 2.0 * u[get_local_id(0)+2] * u[get_local_id(0)+3];
        u[get_local_id(0)+6] = u[get_local_id(0)+2] * u[get_local_id(0)+4]-u[get_local_id(0)+3] * u[get_local_id(0)+5];
        u[get_local_id(0)+7] = u[get_local_id(0)+2] * u[get_local_id(0)+5]+u[get_local_id(0)+3] * u[get_local_id(0)+4];
      }

      if (eightidx < 4)
      {
        i = eightnr*4+eightidx;
        x_zwi[2 * i] = xrcomp[kk+k_zwi];
        x_zwi[2 * i + 1] = xicomp[kk+k_zwi];
      }

      barrier(CLK_LOCAL_MEM_FENCE);

      int wi_0 = (eightidx / 2) * 2;
      int wi_1 = wi_0 + 1;

      int alt_fak = (-1 + ((eightidx % 2)*2));

      int www = ((eightidx / 2) * 2) + (1 - (eightidx % 2));
      c[get_local_id(0)] = u[wi_0+(get_local_id(0)/8)*8]*x_zwi[eightidx+(get_local_id(0)/8)*8] + alt_fak*u[wi_1+(get_local_id(0)/8)*8]*x_zwi[www+(get_local_id(0)/8)*8];

      barrier(CLK_LOCAL_MEM_FENCE);

      d[get_local_id(0)] = v1[eightidx]*c[f1+(eightnr<<3)] + v2[eightidx]*c[f2+(eightnr<<3)];

      barrier(CLK_LOCAL_MEM_FENCE);

      int v3 = 1 - ((eightidx >> 2) << 1);

      w_zwi[get_local_id(0)] = d[(eightidx%4)+(eightnr<<3)] + v3*d[eightidx%4+4+(eightnr<<3)];

      barrier(CLK_LOCAL_MEM_FENCE);
      i = get_local_id(0)-1;
      wrcomp[kk+j_tmp] = w_zwi[i];
      wicomp[kk+j_tmp] = w_zwi[i+1];
    }

  }
}

__kernel void cfft4_opencl_r4_kernel(int is, int n, int ln, int ku, int l_in,
                                     __global double *urcomp, __global double *uicomp,
                                     __global double *xrcomp, __global double *xicomp,
                                     __global double *wrcomp, __global double *wicomp)
{

  const unsigned int gidx = get_local_size(0)*get_group_id(0)+get_local_id(0);

  int eightidx = gidx & 7;

  int eightnr = (get_local_id(0) >> 3);

  int i;
  int l1 = l_in;

  __local int v1[8]; // = {1, 1, 1, 1, 1, 1,-1, 1};
  __local int v2[8]; // = {1, 1,-1,-1, 1, 1, 1,-1};
  if (get_local_id(0) < 8) {
    v1[get_local_id(0)] = 1;
    v2[get_local_id(0)] = 1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) < 2 ) {
    v2[2 + (get_local_id(0) & 1)] = -1;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  if (get_local_id(0) == 0) {
    v1[6] = -1 * is;
    v1[7] *= is;

    v2[6] *= is;
    v2[7] = -1 * is;
  }
  barrier(CLK_LOCAL_MEM_FENCE);

  int n_zwi = n>>2;

  int mask3 = (~7);

  int x0 = eightidx&1;           // 0 1 0 1 0 1 0 1
  int x1 = 1-x0;                 // 1 0 1 0 1 0 1 0
  int x3 = eightidx>>2;          // 0 0 0 0 1 1 1 1
  int x2 = (eightidx>>1)&x3;     // 0 0 0 0 0 0 1 1

  int f1 = x0+(x3<<1)+x2*(-1+(x1<<1));  // 0 1 0 1 2 3 3 2
  int f2 = f1 + 4;                      // 4 5 4 5 6 7 7 6

  int wi_0 = (eightidx >> 1) << 1;
  int wi_1 = wi_0 + 1;

  int v3 = 1 - ((eightidx >> 2) << 1);

  int alt_fak = (-1 + ((int)(eightidx & 1)<<1));

  int www = wi_0 + (1 - (eightidx & 1));

  int jj;
  for (jj=get_group_id(1); jj<ln; jj+=get_num_groups(1))
  {
    int kk;

    __local double w_zwi[BLOCK_SIZE];

    __local double c[BLOCK_SIZE];
    __local double d[BLOCK_SIZE];
    __local double u[BLOCK_SIZE];

    __local double x_tmp[BLOCK_SIZE];

    for( kk = (gidx>>3); kk <= l1 - 1; kk+=((get_num_groups(0)*get_local_size(0))>>3) ) {

      if (eightidx < 2) {
        u[(eightnr<<3)+eightidx] = (double)(1-eightidx);
      }

      barrier(CLK_LOCAL_MEM_FENCE);

      if (eightidx == 0) {
        //u[get_local_id(0)] = 1.0;
        //u[get_local_id(0)+1] = 0.0;

        u[get_local_id(0)+2] = urcomp[kk+ku-1];
        u[get_local_id(0)+3] = is*uicomp[kk+ku-1];
        u[get_local_id(0)+4] = u[get_local_id(0)+2]*u[get_local_id(0)+2] - u[get_local_id(0)+3]*u[get_local_id(0)+3];
        u[get_local_id(0)+5] = 2.0*u[get_local_id(0)+2]*u[get_local_id(0)+3];

        u[get_local_id(0)+6] = u[get_local_id(0)+2]*u[get_local_id(0)+4]-u[get_local_id(0)+3]*u[get_local_id(0)+5];
        u[get_local_id(0)+7] = u[get_local_id(0)+2]*u[get_local_id(0)+5]+u[get_local_id(0)+3]*u[get_local_id(0)+4];
      }

      barrier(CLK_LOCAL_MEM_FENCE);

      if (eightidx < 4) {		
        int k_tmp = (jj*l1) + (eightidx * n_zwi);
        i = (eightnr<<2)+eightidx;
        w_zwi[2 * i    ] = wrcomp[kk+k_tmp];
        w_zwi[2 * i + 1] = wicomp[kk+k_tmp];
      }

      barrier(CLK_LOCAL_MEM_FENCE);
      c[get_local_id(0)] = u[wi_0+(get_local_id(0)&mask3)]*w_zwi[eightidx+(get_local_id(0)&mask3)] + alt_fak*u[wi_1+(get_local_id(0)&mask3)]*w_zwi[www+(get_local_id(0)&mask3)];

      barrier(CLK_LOCAL_MEM_FENCE);

      d[get_local_id(0)] = v1[eightidx]*c[f1+(eightnr<<3)] + v2[eightidx]*c[f2+(eightnr<<3)];

      barrier(CLK_LOCAL_MEM_FENCE);

      x_tmp[get_local_id(0)] = d[(eightidx&3)+(eightnr<<3)] + v3*d[(eightidx&3)+4+(eightnr<<3)];

      barrier(CLK_LOCAL_MEM_FENCE);


      int j_tmp = (jj*l1*4) + l1*(eightidx/2);
      i = get_local_id(0) - 1;
      xrcomp[kk+j_tmp] = x_tmp[i];
      xicomp[kk+j_tmp] = x_tmp[i+1];

    }
  }
}

__kernel void cfft4_opencl_r5_kernel(int is, int l1, int ku,
                                     __global double *urcomp, __global double *uicomp,
                                     __global double *xrcomp, __global double *xicomp,
                                     __global double *wrcomp, __global double *wicomp)
{

  int k;

  const unsigned int gidx = get_local_size(0)*get_group_id(0)+get_local_id(0);

  int eightidx = gidx & (8-1);
  int warp = gidx / WARP_SIZE;
  int warpidx = get_local_id(0) & (WARP_SIZE-1);

  int wcnt = BLOCK_SIZE / WARP_SIZE;

  int eightnr = (get_local_id(0) >> 3);

  int l2 = l1/2;

  int i;
  int t1 = (eightidx>>1);
  int t2 = t1^1;

  __local int jk[4];

  __local double x_zwi[BLOCK_SIZE];

  __local double u_zwi[BLOCK_SIZE];
  __local double c_zwi[BLOCK_SIZE];
  __local double w_zwi[BLOCK_SIZE];

  __local double xxy[2];

  __local double u_in[BLOCK_SIZE/4];

  if (get_local_id(0) == 0)
  {
    jk[0] = 0;
    jk[1] = l1;
    jk[2] = l2;
    jk[3] = l1 + l2;

    xxy[0] = 1.0;
    xxy[1] = (double)is;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  for( k = (gidx>>3); k <= l2 - 1; k+=((get_num_groups(0)*get_local_size(0))>>3) ) {

    if (eightidx < 4)
    {
      i = (eightnr<<2)+eightidx;
      x_zwi[2 * i    ] = xrcomp[k+jk[eightidx]];
      x_zwi[2 * i + 1] = xicomp[k+jk[eightidx]];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx == 0)
    {
      i = eightnr;
      u_in[2 * i] = urcomp[2*k+ku-1];
      u_in[2 * i + 1] = uicomp[2*k+ku-1];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4)
    {
      i = eightnr;
      u_zwi[get_local_id(0)] = t2 + t1*(xxy[eightidx&1]*u_in[2 * i + (eightidx&1)]);
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4) {
      int blaidx = ((eightidx/2)*2)+(1-(eightidx%2));

      double zzwi = u_zwi[(eightnr<<3)+blaidx];

      u_zwi[get_local_id(0)+4] = (1*t2) + t1*((-1 + (2 * (eightidx&1)))*is*zzwi);
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    // 1,0,3,2,5,4,7,6
    int blaidx = ((eightidx/2)*2)+(1-(eightidx%2));
    // -1, 1, -1, 1, -1, 1 ...
    int v2 = (-1 + ((eightidx&1)<<1));
    int v3 = ((eightidx/2)%2);

    double v4 = (double)(v2*v3);

    double xcy1 = u_zwi[(eightnr<<3)+(eightidx/2)*2]*x_zwi[(eightnr<<3)+eightidx];

    double xcy2 = u_zwi[(eightnr<<3)+((eightidx/2)*2+1)]*x_zwi[(eightnr<<3)+blaidx];

    c_zwi[get_local_id(0)] = xcy1 + v4*xcy2;


    barrier(CLK_LOCAL_MEM_FENCE);

    int w1 = (eightidx&1)+((eightidx>>2)<<2);
    int w2 = w1 + 2;
    double w3 = (double)1-(((eightidx/2)%2)*2);

    double x11 = c_zwi[w1+((warp%wcnt)*WARP_SIZE)+(warpidx/8)*8];
    double x22 = c_zwi[w2+((warp%wcnt)*WARP_SIZE)+(warpidx/8)*8];

    w_zwi[get_local_id(0)] = x11 + w3*x22;

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4)
    {
      i = (eightnr<<2)+eightidx;
      wrcomp[k+jk[eightidx]] = w_zwi[2 * i];
      wicomp[k+jk[eightidx]] = w_zwi[2 * i + 1];

    }
  }
}

__kernel void cfft4_opencl_r6_kernel(int is, int l1, int ku,
                                     __global double *urcomp, __global double *uicomp,
                                     __global double *xrcomp, __global double *xicomp,
                                     __global double *wrcomp, __global double *wicomp)
{

  int k;

  const unsigned int gidx = get_global_id(0);

  int eightidx = gidx & (8-1);
  int warp = gidx / WARP_SIZE;
  int warpidx = get_local_id(0) & (WARP_SIZE-1);

  int wcnt = BLOCK_SIZE / WARP_SIZE;

  int eightnr = (get_local_id(0) >> 3);

  int l2 = l1/2;
  int i;

  int t1 = (eightidx>>1);
  int t2 = t1^1;

  __local int jk[4];

  __local double w_zwi[BLOCK_SIZE];

  __local double u_zwi[BLOCK_SIZE];
  __local double c_zwi[BLOCK_SIZE];
  __local double x_zwi[BLOCK_SIZE];

  __local double xxy[2];

  __local double u_in[BLOCK_SIZE/4];

  if (get_local_id(0) == 0) {
    jk[0] = 0;       // 0
    jk[1] = l1;      // 64
    jk[2] = l2;      // 32
    jk[3] = l1 + l2; // 96

    xxy[0] = 1.0;
    xxy[1] = (double)is;
  }

  barrier(CLK_LOCAL_MEM_FENCE);

  for( k = (gidx>>3); k <=l2 - 1; k+=((get_num_groups(0)*get_local_size(0))>>3) )
  {

    if (eightidx < 4)
    {
      i = (eightnr<<2)+eightidx;
      w_zwi[2 * i    ] = wrcomp[k+jk[eightidx]];
      w_zwi[2 * i + 1] = wicomp[k+jk[eightidx]];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx == 0)
    {
      i = eightnr;
      u_in[2 * i] = urcomp[2*k+ku-1];
      u_in[2 * i + 1] = uicomp[2*k+ku-1];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4) {

      i = eightnr;
      u_zwi[get_local_id(0)] = (1*t2) + t1*(xxy[eightidx&1]*u_in[2 * i + (eightidx&1)]);
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4) {

      int blaidx = ((eightidx/2)*2)+(1-(eightidx%2));

      double zzwi = u_zwi[(eightnr<<3)+blaidx];

      u_zwi[get_local_id(0)+4] = (1*t2) + t1*((-1 + (2 * (eightidx&1)))*is*zzwi);
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    // 1,0,3,2,5,4,7,6
    int blaidx = ((eightidx/2)*2)+(1-(eightidx%2));
    // -1, 1, -1, 1, -1, 1 ...
    int v2 = (-1 + ((eightidx&1)<<1));
    int v3 = ((eightidx/2)%2);

    double v4 = (double)(v2*v3);

    double xcy1 = u_zwi[(eightnr<<3)+(eightidx/2)*2]*w_zwi[(eightnr<<3)+eightidx];
    double xcy2 = u_zwi[(eightnr<<3)+((eightidx/2)*2+1)]*w_zwi[(eightnr<<3)+blaidx];

    c_zwi[get_local_id(0)] = xcy1 + v4*xcy2;

    barrier(CLK_LOCAL_MEM_FENCE);

    int w1 = (eightidx&1)+((eightidx>>2)<<2);
    int w2 = w1 + 2;
    double w3 = 1.0 - (((eightidx/2)%2)*2);

    double x11 = c_zwi[w1+((warp%wcnt)*WARP_SIZE)+(warpidx/8)*8];
    double x22 = c_zwi[w2+((warp%wcnt)*WARP_SIZE)+(warpidx/8)*8];

    x_zwi[get_local_id(0)] = x11 + w3*x22;

    barrier(CLK_LOCAL_MEM_FENCE);

    if (eightidx < 4)
    {
      i = (eightnr<<2)+eightidx;
      xrcomp[k+jk[eightidx]] = x_zwi[2 * i];
      xicomp[k+jk[eightidx]] = x_zwi[2 * i + 1];
    }

  }
}
